@@ -1,32 +1,25 @@
 #!/usr/bin/env python3
 # Run three simple jobs in sequence
 import asyncio
-from asyncjobs import Job, Scheduler
+from asyncjobs import Scheduler
+import time
 
 
-# Helper function
-def sleep():
-    import time
-
+def sleep():  # Run in a worker thread by job #2 below
     print(f'{time.ctime()}: Sleep for a second')
     time.sleep(1)
     print(f'{time.ctime()}: Finished sleep')
 
 
+s = Scheduler()
+
 # Job #1 prints uptime
-job1 = Job('#1')
-job1.subprocess_argv = ['uptime']
+s.add_subprocess_job('#1', ['uptime'])
 
 # Job #2 waits for #1 and then sleeps in a thread
-job2 = Job('#2', deps={'#1'})
-job2.thread_func = sleep
+s.add_thread_job('#2', sleep, deps={'#1'})
 
 # Job #3 waits for #2 and then prints uptime (again)
-job3 = Job('#3', deps={'#2'})
-job3.subprocess_argv = ['uptime']
+s.add_subprocess_job('#3', ['uptime'], deps={'#2'})
 
-# Run all jobs in the scheduler
-s = Scheduler()
-for job in [job1, job2, job3]:
-    s.add(job)
 asyncio.run(s.run())
