@@ -1,6 +1,7 @@
 import asyncio
 from contextlib import contextmanager
 from functools import partial
+import itertools
 import logging
 import os
 import pytest
@@ -8,7 +9,7 @@ import signal
 from subprocess import CalledProcessError
 import time
 
-from asyncjobs import basic, external_work
+from verify_events import EventVerifier, ExpectedJobEvents, Whatever
 
 logger = logging.getLogger(__name__)
 
@@ -59,21 +60,7 @@ def assert_elapsed_time_within(time_limit):
         assert after < before + time_limit
 
 
-# Wildcard used for don't-care values in expected event fields
-Whatever = object()
-
-
-def verify_event(expect, actual):
-    expect_keys = set(expect.keys()) - {'MAY_CANCEL'}
-    assert expect_keys == set(actual.keys())
-    keys = sorted(expect_keys)
-    for e, a in zip((expect[k] for k in keys), (actual[k] for k in keys)):
-        if e is not Whatever:
-            assert e == a
-    return True
-
-
-class TSimpleJob(basic.Job):
+class TBasicJob:
     """Async jobs with test instrumentation."""
 
     def __init__(
@@ -88,7 +75,8 @@ class TSimpleJob(basic.Job):
         await_spawn=False,
         result=None,
     ):
-        super().__init__(name=name, deps=deps)
+        self.name = name
+        self.deps = deps
         self.before = set() if before is None else set(before)
         self.call = call
         self.async_sleep = async_sleep
@@ -96,72 +84,67 @@ class TSimpleJob(basic.Job):
         self.await_spawn = await_spawn
         self.result = '{} done'.format(name) if result is None else result
 
-        self._expected_events = []
-        self._expect_event('add'),
-        self._expect_event('start'),
+        self.xevents = ExpectedJobEvents(name)
+        self.xevents.add('add')
+        self.xevents.add('start')
 
-    def _expect_event(self, event, *, may_cancel=False, **kwargs):
-        d = {'event': event, 'job': self.name, 'timestamp': Whatever}
-        d.update(kwargs)
-        if may_cancel:
-            d['MAY_CANCEL'] = True
-        self._expected_events.append(d)
+        # self.deps is processed _before_ .__call__() is invoked
+        if self.deps:
+            if 'MISSING' in self.deps:
+                self.xevents.add('finish', fate='failed')  # expect KeyError
+            else:
+                self.xevents.add(
+                    'await results', jobs=list(self.deps), pending=Whatever
+                )
 
-    async def do_work(self, scheduler):
+    def descendants(self):
+        for spawn in self.spawn:
+            yield spawn
+            yield from spawn.descendants()
+
+    async def do_work(self, ctx):
         result = None
         if self.call:
             result = self.call()
         return self.result if result is None else result
 
-    def expected_events(self):
-        if self._expected_events[-1]['event'] != 'finish':  # we were cancelled
-            self._expect_event('finish', fate='cancelled')
-        return self._expected_events
-
-    async def __call__(self, scheduler):
+    async def __call__(self, ctx):
+        dep_results = ctx.deps
         if self.deps:
-            if self.deps == {'MISSING'}:
-                self._expect_event('finish', fate='failed')  # expect KeyError
-            else:
-                self._expect_event(
-                    'await results', jobs=list(self.deps), pending=Whatever
-                )
-        dep_results = await super().__call__(scheduler)
-        if self.deps:
-            self._expect_event('awaited results')
-        self.logger.debug(f'Results from deps: {dep_results}')
+            self.xevents.add('awaited results')
+        ctx.logger.debug(f'Results from deps: {dep_results}')
 
         if self.async_sleep:
-            self.logger.info(f'Async sleep for {self.async_sleep} seconds…')
+            ctx.logger.info(f'Async sleep for {self.async_sleep} seconds…')
             await asyncio.sleep(self.async_sleep)
-            self.logger.info('Finished async sleep')
+            ctx.logger.info('Finished async sleep')
 
-        result = await self.do_work(scheduler)
+        result = await self.do_work(ctx)
 
         for job in self.spawn:
-            scheduler.add(job)
+            ctx.add_job(job.name, job, job.deps)
 
         if self.await_spawn and self.spawn:
             spawn = [job.name for job in self.spawn]
-            self._expect_event('await results', jobs=spawn, pending=spawn)
-            await scheduler.results(*[job.name for job in self.spawn])
-            self._expect_event('awaited results')
+            self.xevents.add('await results', jobs=spawn, pending=spawn)
+            await ctx.results(*[job.name for job in self.spawn])
+            self.xevents.add('awaited results')
 
         for b in self.before:
-            assert b in scheduler.tasks  # The other job has been started
-            assert not scheduler.tasks[b].done()  # but is not yet finished
+            assert b in ctx._scheduler.tasks  # The other job has started
+            assert not ctx._scheduler.tasks[b].done()  # but not yet finished
 
         if isinstance(result, Exception):
-            self.logger.info(f'Raising exception: {result}')
-            self._expect_event('finish', fate='failed')
+            ctx.logger.info(f'Raising exception: {result}')
+            self.xevents.add('finish', fate='failed')
             raise result
         else:
-            self.logger.info(f'Returning result: {result!r}')
-            self._expect_event('finish', fate='success')
+            ctx.logger.info(f'Returning result: {result!r}')
+            self.xevents.add('finish', fate='success')
             return result
 
 
-class TExternalWorkJob(TSimpleJob, external_work.Job):
+class TExternalWorkJob(TBasicJob):
     """Test jobs with thread/subprocess capabilities."""
 
     def __init__(
@@ -177,7 +160,7 @@ class TExternalWorkJob(TSimpleJob, external_work.Job):
         if thread_sleep:
             if thread is not None:
                 raise ValueError('Cannot both sleep and work in thread')
-            self.thread = partial(time.sleep, thread_sleep)
+            self.thread = lambda ctx: time.sleep(thread_sleep)
         else:
             self.thread = thread
         if subproc_sleep:
@@ -187,136 +170,58 @@ class TExternalWorkJob(TSimpleJob, external_work.Job):
         else:
             self.subproc = subproc
 
-    async def _do_thread_stuff(self, scheduler):
-        self.logger.debug(f'Await call {self.thread} in thread…')
-        self._expect_event('await worker slot')
-        self._expect_event('awaited worker slot', may_cancel=True)
-        self._expect_event(
-            'await worker thread', may_cancel=True, func=Whatever
-        )
+    async def _do_thread_stuff(self, ctx):
+        ctx.logger.debug(f'Await call {self.thread} in thread…')
+        self.xevents.add('await worker slot')
+        self.xevents.add('awaited worker slot', may_cancel=True)
+        self.xevents.add('await worker thread', may_cancel=True, func=Whatever)
         try:
-            ret = await self.call_in_thread(self.thread)
-            self._expect_event('awaited worker thread', fate='success')
+            ret = await ctx.call_in_thread(self.thread, ctx)
+            self.xevents.add('awaited worker thread', fate='success')
         except asyncio.CancelledError:
-            self._expect_event(
+            self.xevents.add(
                 'awaited worker thread', may_cancel=True, fate='cancelled',
             )
             raise
         except Exception as e:
             ret = e
-            self._expect_event('awaited worker thread', fate='failed')
-        self.logger.debug(f'Finished thread call: {ret!r}')
+            self.xevents.add('awaited worker thread', fate='failed')
+        ctx.logger.debug(f'Finished thread call: {ret!r}')
         return ret
 
-    async def _do_subproc_stuff(self, scheduler):
-        self.logger.debug(f'Await run {self.subproc} in subprocess…')
-        self._expect_event('await worker slot')
-        self._expect_event('awaited worker slot', may_cancel=True)
-        self._expect_event(
+    async def _do_subproc_stuff(self, ctx):
+        ctx.logger.debug(f'Await run {self.subproc} in subprocess…')
+        self.xevents.add('await worker slot')
+        self.xevents.add('awaited worker slot', may_cancel=True)
+        self.xevents.add(
             'await worker proc', may_cancel=True, argv=self.subproc
         )
         try:
-            ret = await self.run_in_subprocess(self.subproc)
-            self._expect_event('awaited worker proc', exit=0)
+            ret = await ctx.run_in_subprocess(self.subproc)
+            self.xevents.add('awaited worker proc', returncode=0)
         except asyncio.CancelledError:
-            self._expect_event(
-                'awaited worker proc', may_cancel=True, exit=-15
+            self.xevents.add(
+                'awaited worker proc', may_cancel=True, returncode=-15
             )
             raise
         except Exception as e:
             ret = e
             if isinstance(e, CalledProcessError):
-                self._expect_event('awaited worker proc', exit=e.returncode)
-        self.logger.debug(f'Finished subprocess run: {ret}')
+                self.xevents.add(
+                    'awaited worker proc', returncode=e.returncode
+                )
+        ctx.logger.debug(f'Finished subprocess run: {ret}')
         return ret
 
-    async def do_work(self, scheduler):
-        result = await super().do_work(scheduler)
+    async def do_work(self, ctx):
+        result = await super().do_work(ctx)
         if self.thread:
-            result = await self._do_thread_stuff(scheduler)
+            result = await self._do_thread_stuff(ctx)
         if self.subproc:
-            result = await self._do_subproc_stuff(scheduler)
+            result = await self._do_subproc_stuff(ctx)
         return result
 
 
-def verify_events(scheduler, start_time, actual, todo):
-    after = time.time()
-    num_jobs = len(todo)
-    num_tasks = len(scheduler.tasks)
-    expect = {j.name: j.expected_events() for j in todo}
-
-    # Timestamps are in sorted order and all between 'start_time' and 'after'
-    timestamps = [e['timestamp'] for e in actual]
-    assert timestamps == sorted(timestamps)
-    assert timestamps[0] >= start_time
-    assert timestamps[-1] <= after
-
-    # Initial jobs are added before execution starts
-    expect_adds = [e.pop(0) for e in expect.values()]
-    for e, a in zip(expect_adds, actual[:num_jobs]):
-        assert verify_event(e, a)
-    actual = actual[num_jobs:]
-
-    # Overall execution start and finish
-    overall_start, overall_finish = actual.pop(0), actual.pop()
-    assert verify_event(
-        {
-            'event': 'start',
-            'num_jobs': num_jobs,
-            'keep_going': Whatever,
-            'timestamp': Whatever,
-        },
-        overall_start,
-    )
-    assert verify_event(
-        {'event': 'finish', 'num_tasks': num_tasks, 'timestamp': Whatever},
-        overall_finish,
-    )
-
-    if expect:
-        # Jobs are started
-        expect_starts = [e.pop(0) for e in expect.values()]
-        for e, a in zip(expect_starts, actual[:num_jobs]):
-            assert verify_event(e, a)
-        actual = actual[num_jobs:]
-
-        # Await task execution
-        await_tasks, awaited_tasks = actual.pop(0), actual.pop()
-        assert verify_event(
-            {
-                'event': 'await tasks',
-                'jobs': [j.name for j in todo],
-                'timestamp': Whatever,
-            },
-            await_tasks,
-        )
-        assert verify_event(
-            {'event': 'awaited tasks', 'timestamp': Whatever}, awaited_tasks,
-        )
-
-        # Remaining events belong to individual tasks
-        # This includes expected events from tasks that were spawned from
-        # other tasks (hence not part of the original 'todo').
-        for name, job in scheduler.jobs.items():
-            if name not in expect:  # job was spawned after .run()
-                expect[name] = job.expected_events()
-        # Verify each event by pairing it w/that job's next expected event
-        while actual:
-            a = actual.pop(0)
-            job_name = a['job']
-            e = expect[job_name].pop(0)
-            if a['event'] == 'finish' and a['fate'] == 'cancelled':
-                while e.get('MAY_CANCEL', False):
-                    e = expect[job_name].pop(0)
-            assert verify_event(e, a)
-
-        # No more expected events
-        assert all(len(e) == 0 for e in expect.values())
-
-    assert not actual  # no more actual events
-    return True
-
-
 @pytest.fixture(params=[1, 2, 4, 100])
 def num_workers(request):
     return request.param
@@ -335,16 +240,26 @@ def scheduler_with_workers(num_workers):
 
 
 @contextmanager
-def setup_scheduler(scheduler_class, todo):
-    events = []
-    scheduler = scheduler_class(event_handler=events.append)
-    before = time.time()
-    for job in todo:
-        scheduler.add(job)
-    yield scheduler
-    verify_events(scheduler, before, events, todo)
+def verified_events(scheduler, todo):
+    """Check events generated by scheduler against expectations.
+
+    Verify that the given scheduler generates the expected events while in
+    this context. We set the scheduler's event_handler to collect all events
+    that occur while in the context, and upon exiting the context, we verify
+    the events emitted by the scheduler against the events expected by the
+    TBasicJob instances in 'todo'.
+    """
+    initial_xevents = [j.xevents for j in todo]
+    spawned_xevents = [
+        j.xevents for j in itertools.chain(*[j.descendants() for j in todo])
+    ]
+    ev = EventVerifier()
+    scheduler.event_handler = ev
+    yield
+    ev.verify_all(initial_xevents, spawned_xevents)
 
 
+# Used to signal the expectation of a cancelled task
 Cancelled = object()
 
 
@@ -403,7 +318,9 @@ def verify_output(capfd):
             else:  # no match for any stream
                 assert False, f'actual: {actual_line!r}, expect: {expected!r}'
         # no more lines expected
-        assert all(e == [] for e in expect_lines_from_streams)
+        for e in expect_lines_from_streams:
+            if len(e):
+                assert False, f'expected lines not found: {e!r}'
 
     def _verify_output(expect_stdout_streams, expect_stderr_streams=None):
         actual = capfd.readouterr()
